@@ -10,8 +10,8 @@ use std::{
 use serde::{de, Deserialize, Deserializer};
 
 use crate::{
-    ipfilter::IpFilter, KeySet, NoCipher, NtpClock, NtpPacket, NtpTimestamp, PacketParsingError,
-    SystemSnapshot,
+    ipfilter::IpFilter, KeySet, NoCipher, NtpClock, NtpPacket, NtpTimestamp, NtpVersion,
+    PacketParsingError, SystemSnapshot,
 };
 
 pub enum ServerAction<'a> {
@@ -79,6 +79,7 @@ pub struct ServerConfig {
     pub rate_limiting_cache_size: usize,
     pub rate_limiting_cutoff: Duration,
     pub require_nts: Option<FilterAction>,
+    pub accepted_versions: Vec<NtpVersion>,
 }
 
 pub struct Server<C> {
@@ -209,12 +210,29 @@ impl<C: NtpClock> Server<C> {
 
         // Generate the appropriate response
         let version = packet.version();
+
+        if !self.config.accepted_versions.contains(&version) {
+            // handle this packet as if we don't know it
+            stats_handler.register(
+                version.as_u8(),
+                false,
+                ServerReason::Policy,
+                ServerResponse::Ignore,
+            );
+            return ServerAction::Ignore;
+        }
+
         let nts = cookie.is_some() || action == ServerResponse::NTSNak;
 
         // ignore non-NTS packets when configured to require NTS
         if let (false, Some(non_nts_action)) = (nts, self.config.require_nts) {
             if non_nts_action == FilterAction::Ignore {
-                stats_handler.register(version, nts, ServerReason::Policy, ServerResponse::Ignore);
+                stats_handler.register(
+                    version.into(),
+                    nts,
+                    ServerReason::Policy,
+                    ServerResponse::Ignore,
+                );
                 return ServerAction::Ignore;
             } else {
                 action = ServerResponse::Deny;
@@ -262,7 +280,7 @@ impl<C: NtpClock> Server<C> {
         };
         match result {
             Ok(_) => {
-                stats_handler.register(version, nts, reason, action);
+                stats_handler.register(version.into(), nts, reason, action);
                 let length = cursor.position();
                 ServerAction::Respond {
                     message: &cursor.into_inner()[..length as _],
@@ -271,7 +289,7 @@ impl<C: NtpClock> Server<C> {
             Err(e) => {
                 tracing::error!("Could not serialize response: {}", e);
                 stats_handler.register(
-                    version,
+                    version.into(),
                     nts,
                     ServerReason::InternalError,
                     ServerResponse::Ignore,
@@ -313,13 +331,9 @@ impl<T: std::hash::Hash + Eq> TimestampedCache<T> {
     }
 
     fn index(&self, item: &T) -> usize {
-        use std::hash::{BuildHasher, Hasher};
+        use std::hash::BuildHasher;
 
-        let mut hasher = self.randomstate.build_hasher();
-
-        item.hash(&mut hasher);
-
-        hasher.finish() as usize % self.elements.len()
+        self.randomstate.hash_one(item) as usize % self.elements.len()
     }
 
     fn is_allowed(&mut self, item: T, timestamp: Instant, cutoff: Duration) -> bool {
@@ -411,8 +425,6 @@ impl<'de> Deserialize<'de> for IpSubnet {
 mod tests {
     use std::net::{Ipv4Addr, Ipv6Addr};
 
-    use serde_test::{assert_de_tokens, assert_de_tokens_error, Token};
-
     use crate::{
         nts_record::AeadAlgorithm, packet::AesSivCmac256, Cipher, DecodedServerCookie,
         KeySetProvider, NtpDuration, NtpLeapIndicator, PollIntervalLimits,
@@ -513,6 +525,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_secs(1),
             rate_limiting_cache_size: 0,
             require_nts: None,
+            accepted_versions: vec![NtpVersion::V4],
         };
         let clock = TestClock {
             cur: NtpTimestamp::from_fixed_int(200),
@@ -583,6 +596,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_secs(1),
             rate_limiting_cache_size: 0,
             require_nts: None,
+            accepted_versions: vec![NtpVersion::V4],
         };
         server.update_config(config);
 
@@ -621,6 +635,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_secs(1),
             rate_limiting_cache_size: 0,
             require_nts: None,
+            accepted_versions: vec![NtpVersion::V4],
         };
         let clock = TestClock {
             cur: NtpTimestamp::from_fixed_int(200),
@@ -697,6 +712,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_secs(1),
             rate_limiting_cache_size: 0,
             require_nts: None,
+            accepted_versions: vec![NtpVersion::V4],
         };
         server.update_config(config);
 
@@ -729,6 +745,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_millis(100),
             rate_limiting_cache_size: 32,
             require_nts: None,
+            accepted_versions: vec![NtpVersion::V4],
         };
         let clock = TestClock {
             cur: NtpTimestamp::from_fixed_int(200),
@@ -829,6 +846,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_millis(100),
             rate_limiting_cache_size: 0,
             require_nts: None,
+            accepted_versions: vec![NtpVersion::V4],
         };
 
         server.update_config(config);
@@ -904,6 +922,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_millis(100),
             rate_limiting_cache_size: 0,
             require_nts: None,
+            accepted_versions: vec![NtpVersion::V4],
         };
         let clock = TestClock {
             cur: NtpTimestamp::from_fixed_int(200),
@@ -967,6 +986,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_millis(100),
             rate_limiting_cache_size: 0,
             require_nts: None,
+            accepted_versions: vec![NtpVersion::V4],
         };
         server.update_config(config);
 
@@ -996,6 +1016,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_millis(100),
             rate_limiting_cache_size: 0,
             require_nts: None,
+            accepted_versions: vec![NtpVersion::V4],
         };
         server.update_config(config);
 
@@ -1025,6 +1046,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_millis(100),
             rate_limiting_cache_size: 0,
             require_nts: None,
+            accepted_versions: vec![NtpVersion::V4],
         };
         server.update_config(config);
 
@@ -1054,6 +1076,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_millis(100),
             rate_limiting_cache_size: 0,
             require_nts: None,
+            accepted_versions: vec![NtpVersion::V4],
         };
         server.update_config(config);
 
@@ -1086,6 +1109,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_millis(100),
             rate_limiting_cache_size: 0,
             require_nts: Some(FilterAction::Ignore),
+            accepted_versions: vec![NtpVersion::V4],
         };
         let clock = TestClock {
             cur: NtpTimestamp::from_fixed_int(200),
@@ -1176,6 +1200,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_secs(1),
             rate_limiting_cache_size: 0,
             require_nts: Some(FilterAction::Ignore),
+            accepted_versions: vec![NtpVersion::V4],
         };
         let clock = TestClock {
             cur: NtpTimestamp::from_fixed_int(200),
@@ -1260,7 +1285,6 @@ mod tests {
         assert!(packet.is_kiss_deny());
     }
 
-    #[cfg(feature = "ntpv5")]
     #[test]
     fn test_server_v5() {
         let config = ServerConfig {
@@ -1275,6 +1299,7 @@ mod tests {
             rate_limiting_cutoff: Duration::from_millis(100),
             rate_limiting_cache_size: 0,
             require_nts: None,
+            accepted_versions: vec![NtpVersion::V5],
         };
         let clock = TestClock {
             cur: NtpTimestamp::from_fixed_int(200),
@@ -1340,6 +1365,86 @@ mod tests {
         assert!(packet.is_kiss_deny());
     }
 
+    #[test]
+    fn test_server_ignore_version() {
+        let config = ServerConfig {
+            denylist: FilterList {
+                filter: vec![],
+                action: FilterAction::Deny,
+            },
+            allowlist: FilterList {
+                filter: vec!["0.0.0.0/0".parse().unwrap()],
+                action: FilterAction::Ignore,
+            },
+            rate_limiting_cutoff: Duration::from_millis(1000),
+            rate_limiting_cache_size: 0,
+            require_nts: None,
+            accepted_versions: vec![NtpVersion::V3, NtpVersion::V4],
+        };
+        let clock = TestClock {
+            cur: NtpTimestamp::from_fixed_int(200),
+        };
+        let mut stats = TestStatHandler::default();
+
+        let mut server = Server::new(
+            config,
+            clock,
+            SystemSnapshot::default(),
+            KeySetProvider::new(1).get(),
+        );
+
+        let (packet, _) = NtpPacket::poll_message_v5(PollIntervalLimits::default().min);
+        let serialized = serialize_packet_unencrypted(&packet);
+
+        let mut buf = [0; 1024];
+        let response = server.handle(
+            "128.0.0.1".parse().unwrap(),
+            NtpTimestamp::from_fixed_int(100),
+            &serialized,
+            &mut buf,
+            &mut stats,
+        );
+
+        assert_eq!(
+            stats.last_register.take(),
+            Some((5, false, ServerReason::Policy, ServerResponse::Ignore))
+        );
+        assert!(matches!(response, ServerAction::Ignore));
+
+        server.update_config(ServerConfig {
+            denylist: FilterList {
+                filter: vec![],
+                action: FilterAction::Deny,
+            },
+            allowlist: FilterList {
+                filter: vec!["0.0.0.0/0".parse().unwrap()],
+                action: FilterAction::Ignore,
+            },
+            rate_limiting_cutoff: Duration::from_millis(100),
+            rate_limiting_cache_size: 0,
+            require_nts: None,
+            accepted_versions: vec![NtpVersion::V5],
+        });
+
+        let (packet, _) = NtpPacket::poll_message(PollIntervalLimits::default().min);
+        let serialized = serialize_packet_unencrypted(&packet);
+
+        let mut buf = [0; 1024];
+        let response = server.handle(
+            "128.0.0.1".parse().unwrap(),
+            NtpTimestamp::from_fixed_int(100),
+            &serialized,
+            &mut buf,
+            &mut stats,
+        );
+
+        assert_eq!(
+            stats.last_register.take(),
+            Some((4, false, ServerReason::Policy, ServerResponse::Ignore))
+        );
+        assert!(matches!(response, ServerAction::Ignore));
+    }
+
     // TimestampedCache tests
     #[test]
     fn timestamped_cache() {
@@ -1376,9 +1481,16 @@ mod tests {
     fn test_ipv4_subnet_parse() {
         use std::str::FromStr;
 
-        assert!(IpSubnet::from_str("bla/5").is_err());
-        assert!(IpSubnet::from_str("0.0.0.0").is_err());
-        assert!(IpSubnet::from_str("0.0.0.0/33").is_err());
+        assert!(matches!(
+            IpSubnet::from_str("bla/5"),
+            Err(SubnetParseError::Ip(_))
+        ));
+        assert_eq!(IpSubnet::from_str("0.0.0.0"), Err(SubnetParseError::Subnet));
+        assert_eq!(
+            IpSubnet::from_str("0.0.0.0/33"),
+            Err(SubnetParseError::Mask)
+        );
+
         assert_eq!(
             IpSubnet::from_str("0.0.0.0/0"),
             Ok(IpSubnet {
@@ -1394,25 +1506,20 @@ mod tests {
             })
         );
 
-        assert_de_tokens_error::<IpSubnet>(
-            &[Token::Str("bla/5")],
-            "invalid IP address syntax in subnet",
-        );
-        assert_de_tokens_error::<IpSubnet>(&[Token::Str("0.0.0.0")], "Invalid subnet syntax");
-        assert_de_tokens_error::<IpSubnet>(&[Token::Str("0.0.0.0/33")], "Invalid subnet mask");
-        assert_de_tokens(
-            &IpSubnet {
+        assert_eq!(
+            serde_json::from_str::<IpSubnet>(r#""0.0.0.0/0""#).unwrap(),
+            IpSubnet {
                 addr: IpAddr::V4(Ipv4Addr::UNSPECIFIED),
                 mask: 0,
-            },
-            &[Token::Str("0.0.0.0/0")],
+            }
         );
-        assert_de_tokens(
-            &IpSubnet {
+
+        assert_eq!(
+            serde_json::from_str::<IpSubnet>(r#""127.0.0.1/32""#).unwrap(),
+            IpSubnet {
                 addr: IpAddr::V4(Ipv4Addr::LOCALHOST),
                 mask: 32,
-            },
-            &[Token::Str("127.0.0.1/32")],
+            }
         );
     }
 
@@ -1420,9 +1527,13 @@ mod tests {
     fn test_ipv6_subnet_parse() {
         use std::str::FromStr;
 
-        assert!(IpSubnet::from_str("bla/5").is_err());
-        assert!(IpSubnet::from_str("::").is_err());
-        assert!(IpSubnet::from_str("::/129").is_err());
+        assert!(matches!(
+            IpSubnet::from_str("bla/5"),
+            Err(SubnetParseError::Ip(_))
+        ));
+        assert_eq!(IpSubnet::from_str("::"), Err(SubnetParseError::Subnet));
+        assert_eq!(IpSubnet::from_str("::/129"), Err(SubnetParseError::Mask));
+
         assert_eq!(
             IpSubnet::from_str("::/0"),
             Ok(IpSubnet {
@@ -1438,25 +1549,20 @@ mod tests {
             })
         );
 
-        assert_de_tokens_error::<IpSubnet>(
-            &[Token::Str("bla/5")],
-            "invalid IP address syntax in subnet",
-        );
-        assert_de_tokens_error::<IpSubnet>(&[Token::Str("::")], "Invalid subnet syntax");
-        assert_de_tokens_error::<IpSubnet>(&[Token::Str("::/129")], "Invalid subnet mask");
-        assert_de_tokens(
-            &IpSubnet {
+        assert_eq!(
+            serde_json::from_str::<IpSubnet>(r#""::/0""#).unwrap(),
+            IpSubnet {
                 addr: IpAddr::V6(Ipv6Addr::UNSPECIFIED),
                 mask: 0,
-            },
-            &[Token::Str("::/0")],
+            }
         );
-        assert_de_tokens(
-            &IpSubnet {
+
+        assert_eq!(
+            serde_json::from_str::<IpSubnet>(r#""::1/128""#).unwrap(),
+            IpSubnet {
                 addr: IpAddr::V6(Ipv6Addr::LOCALHOST),
                 mask: 128,
-            },
-            &[Token::Str("::1/128")],
+            }
         );
     }
 }
