@@ -5,40 +5,63 @@ use std::sync::Arc;
 use std::time::Duration;
 use std::{fmt::Debug, hash::Hash};
 
-#[cfg(feature = "ntpv5")]
 use crate::packet::v5::server_reference_id::{BloomFilter, ServerId};
 use crate::source::{NtpSourceUpdate, SourceSnapshot};
 use crate::{
     algorithm::{StateUpdate, TimeSyncController},
     clock::NtpClock,
-    config::{SourceDefaultsConfig, SynchronizationConfig},
+    config::{SourceConfig, SynchronizationConfig},
     identifiers::ReferenceId,
     packet::NtpLeapIndicator,
     source::{NtpSource, NtpSourceActionIterator, ProtocolVersion, SourceNtsData},
     time_types::NtpDuration,
 };
-use crate::{OneWaySource, OneWaySourceUpdate};
+use crate::{NtpTimestamp, OneWaySource, OneWaySourceUpdate};
 
-#[derive(Debug, Clone, Copy, Serialize, Deserialize, PartialEq, Eq)]
+#[derive(Debug, Clone, Copy, Serialize, Deserialize, PartialEq)]
 pub struct TimeSnapshot {
     /// Precision of the local clock
     pub precision: NtpDuration,
     /// Current root delay
     pub root_delay: NtpDuration,
-    /// Current root dispersion
-    pub root_dispersion: NtpDuration,
+    /// t=0 for root variance calculation
+    pub root_variance_base_time: NtpTimestamp,
+    /// Constant contribution for root variance
+    pub root_variance_base: f64,
+    /// Linear (*t) contribution for root variance
+    pub root_variance_linear: f64,
+    /// Quadratic (*t*t) contribution for root variance
+    pub root_variance_quadratic: f64,
+    /// Cubic (*t*t*t) contribution for root variance
+    pub root_variance_cubic: f64,
     /// Current leap indicator state
     pub leap_indicator: NtpLeapIndicator,
     /// Total amount that the clock has stepped
     pub accumulated_steps: NtpDuration,
 }
 
+impl TimeSnapshot {
+    pub fn root_dispersion(&self, now: NtpTimestamp) -> NtpDuration {
+        let t = (now - self.root_variance_base_time).to_seconds();
+        NtpDuration::from_seconds(
+            self.root_variance_base
+                + t * self.root_variance_linear
+                + t.powi(2) * self.root_variance_quadratic
+                + t.powi(3) * self.root_variance_cubic,
+        )
+    }
+}
+
 impl Default for TimeSnapshot {
     fn default() -> Self {
         Self {
             precision: NtpDuration::from_exponent(-18),
             root_delay: NtpDuration::ZERO,
-            root_dispersion: NtpDuration::ZERO,
+            root_variance_base_time: NtpTimestamp::default(),
+            root_variance_base: 0.0,
+            root_variance_linear: 0.0,
+            root_variance_quadratic: 0.0,
+            root_variance_cubic: 0.0,
             leap_indicator: NtpLeapIndicator::Unknown,
             accumulated_steps: NtpDuration::ZERO,
         }
@@ -56,12 +79,9 @@ pub struct SystemSnapshot {
     /// Timekeeping data
     #[serde(flatten)]
     pub time_snapshot: TimeSnapshot,
-
-    #[cfg(feature = "ntpv5")]
     /// Bloom filter that contains all currently used time sources
     #[serde(skip)]
     pub bloom_filter: BloomFilter,
-    #[cfg(feature = "ntpv5")]
     /// NTPv5 reference ID for this instance
     #[serde(skip)]
     pub server_id: ServerId,
@@ -85,20 +105,17 @@ impl SystemSnapshot {
             self.reference_id = source_id;
         }
 
-        #[cfg(feature = "ntpv5")]
-        {
-            self.bloom_filter = BloomFilter::new();
-            for source in used_sources {
-                if let SourceSnapshot::Ntp(source) = source {
-                    if let Some(bf) = &source.bloom_filter {
-                        self.bloom_filter.add(bf);
-                    } else if let ProtocolVersion::V5 = source.protocol_version {
-                        tracing::warn!("Using NTPv5 source without a bloom filter!");
-                    }
+        self.bloom_filter = BloomFilter::new();
+        for source in used_sources {
+            if let SourceSnapshot::Ntp(source) = source {
+                if let Some(bf) = &source.bloom_filter {
+                    self.bloom_filter.add(bf);
+                } else if let ProtocolVersion::V5 = source.protocol_version {
+                    tracing::warn!("Using NTPv5 source without a bloom filter!");
                 }
             }
-            self.bloom_filter.add_id(&self.server_id);
         }
+        self.bloom_filter.add_id(&self.server_id);
     }
 }
 
@@ -109,10 +126,8 @@ impl Default for SystemSnapshot {
             reference_id: ReferenceId::NONE,
             accumulated_steps_threshold: None,
             time_snapshot: TimeSnapshot::default(),
-            #[cfg(feature = "ntpv5")]
             bloom_filter: BloomFilter::new(),
-            #[cfg(feature = "ntpv5")]
-            server_id: ServerId::new(&mut rand::thread_rng()),
+            server_id: ServerId::default(),
         }
     }
 }
@@ -185,7 +200,6 @@ macro_rules! actions {
 
 pub struct System<SourceId, Controller> {
     synchronization_config: SynchronizationConfig,
-    source_defaults_config: SourceDefaultsConfig,
     system: SystemSnapshot,
     ip_list: Arc<[IpAddr]>,
 
@@ -201,7 +215,6 @@ impl<SourceId: Hash + Eq + Copy + Debug, Controller: TimeSyncController<SourceId
     pub fn new(
         clock: Controller::Clock,
         synchronization_config: SynchronizationConfig,
-        source_defaults_config: SourceDefaultsConfig,
         algorithm_config: Controller::AlgorithmConfig,
         ip_list: Arc<[IpAddr]>,
     ) -> Result<Self, <Controller::Clock as NtpClock>::Error> {
@@ -214,20 +227,16 @@ impl<SourceId: Hash + Eq + Copy + Debug, Controller: TimeSyncController<SourceId
         if synchronization_config.local_stratum == 1 {
             // We are a stratum 1 server so mark our selves synchronized.
             system.time_snapshot.leap_indicator = NtpLeapIndicator::NoWarning;
+            // Set the reference id for the system
+            system.reference_id = synchronization_config.reference_id.to_reference_id();
         }
 
         Ok(System {
             synchronization_config,
-            source_defaults_config,
             system,
             ip_list,
             sources: Default::default(),
-            controller: Controller::new(
-                clock,
-                synchronization_config,
-                source_defaults_config,
-                algorithm_config,
-            )?,
+            controller: Controller::new(clock, synchronization_config, algorithm_config)?,
             controller_took_control: false,
         })
     }
@@ -251,15 +260,37 @@ impl<SourceId: Hash + Eq + Copy + Debug, Controller: TimeSyncController<SourceId
     pub fn create_sock_source(
         &mut self,
         id: SourceId,
+        source_config: SourceConfig,
         measurement_noise_estimate: f64,
     ) -> Result<
         OneWaySource<Controller::OneWaySourceController>,
         <Controller::Clock as NtpClock>::Error,
     > {
         self.ensure_controller_control()?;
-        let controller = self
-            .controller
-            .add_one_way_source(id, measurement_noise_estimate);
+        let controller =
+            self.controller
+                .add_one_way_source(id, source_config, measurement_noise_estimate, None);
+        self.sources.insert(id, None);
+        Ok(OneWaySource::new(controller))
+    }
+
+    pub fn create_pps_source(
+        &mut self,
+        id: SourceId,
+        source_config: SourceConfig,
+        measurement_noise_estimate: f64,
+        period: f64,
+    ) -> Result<
+        OneWaySource<Controller::OneWaySourceController>,
+        <Controller::Clock as NtpClock>::Error,
+    > {
+        self.ensure_controller_control()?;
+        let controller = self.controller.add_one_way_source(
+            id,
+            source_config,
+            measurement_noise_estimate,
+            Some(period),
+        );
         self.sources.insert(id, None);
         Ok(OneWaySource::new(controller))
     }
@@ -268,6 +299,7 @@ impl<SourceId: Hash + Eq + Copy + Debug, Controller: TimeSyncController<SourceId
     pub fn create_ntp_source(
         &mut self,
         id: SourceId,
+        source_config: SourceConfig,
         source_addr: SocketAddr,
         protocol_version: ProtocolVersion,
         nts: Option<Box<SourceNtsData>>,
@@ -279,11 +311,11 @@ impl<SourceId: Hash + Eq + Copy + Debug, Controller: TimeSyncController<SourceId
         <Controller::Clock as NtpClock>::Error,
     > {
         self.ensure_controller_control()?;
-        let controller = self.controller.add_source(id);
+        let controller = self.controller.add_source(id, source_config);
         self.sources.insert(id, None);
         Ok(NtpSource::new(
             source_addr,
-            self.source_defaults_config,
+            source_config,
             protocol_version,
             controller,
             nts,
@@ -412,8 +444,7 @@ mod tests {
                     reach: Default::default(),
                     stratum: 2,
                     reference_id: ReferenceId::NONE,
-                    protocol_version: Default::default(),
-                    #[cfg(feature = "ntpv5")]
+                    protocol_version: ProtocolVersion::v4_upgrading_to_v5_with_default_tries(),
                     bloom_filter: None,
                 }),
                 SourceSnapshot::Ntp(NtpSourceSnapshot {
@@ -423,8 +454,7 @@ mod tests {
                     reach: Default::default(),
                     stratum: 3,
                     reference_id: ReferenceId::NONE,
-                    protocol_version: Default::default(),
-                    #[cfg(feature = "ntpv5")]
+                    protocol_version: ProtocolVersion::v4_upgrading_to_v5_with_default_tries(),
                     bloom_filter: None,
                 }),
             ]
